@@ -1,19 +1,33 @@
 import sqlalchemy as sa
 from sqlalchemy_utils.functions import create_database
-from pyg_base import cache, cfg_read, as_list, dictable, replace, Dict, is_dict, is_dictable, is_strs, is_str, is_int, is_date, dt2str, ulist, try_back, unique
+from pyg_base import cache, cfg_read, as_list, dictable, lower, replace, Dict, is_dict, is_dictable, is_strs, is_str, is_int, is_date, dt2str, ulist, try_back, unique
 from pyg_encoders import as_reader, as_writer, dumps, loads
 from sqlalchemy import Table, Column, Integer, String, MetaData, Identity, Float, DATE, DATETIME, TIME, select, func, not_, desc, asc
 from sqlalchemy.orm import Session
+from sqlalchemy.types import NUMERIC, FLOAT, INT
 import datetime
 from copy import copy
 from pyg_base import logger
 from functools import partial
+import pandas as pd
 
 _id = '_id'
 _doc = 'doc'
 _root = 'root'
 _deleted = 'deleted'
 _archived = 'archived_'
+_pd_is_old = pd.__version__.startswith('0')
+
+_types = {str: String, 'str' : String, 
+          int : Integer, 'int' : Integer,
+          float: Float, 
+          'dec' : sa.DECIMAL(0), 'dec1' : sa.DECIMAL(1), 'dec2' : sa.DECIMAL(2), 'dec3' : sa.DECIMAL(3), 'dec4' : sa.DECIMAL(4), 'dec5' : sa.DECIMAL(5), 
+          datetime.date: DATE, 'date' : DATE,
+          datetime.datetime : DATETIME, 'datetime' : DATETIME,
+          datetime.time: TIME, 'time' : TIME,
+          bin : sa.VARBINARY}
+
+_orders = {1 : asc, True: asc, 'asc': asc, asc : asc, -1: desc, False: desc, 'desc': desc, desc: desc}
 
 @cache
 def _servers():
@@ -28,10 +42,33 @@ def _servers():
 def _schema(schema = None):
     return schema or cfg_read().get('sql_schema', 'dbo')
 
+
+@cache
+def _databases():
+    """
+    configures the sql_databases
+    from pyg import * 
+    
+    cfg = cfg_read()
+    cfg['sql_database'] = {None : 'db', 'res' : 'res'}
+    cfg_write(cfg)
+
+    """
+    dbs = cfg_read().get('sql_database', {None: 'master'})
+    if isinstance(dbs, str):
+        dbs = {dbs : dbs}
+    if None not in dbs:
+        dbs[None] = dbs.get('null')
+    dbs[True] = dbs.get('true')
+    dbs[False] = dbs.get('false')
+    return dbs
+
 @cache
 def _database(db = None):
-    return db or cfg_read().get('sql_database', 'master')
-
+    dbs = _databases()
+    db = dbs.get(db, db)
+    db = dbs.get(db, db)
+    return db
     
 def get_server(server = None):
     """
@@ -89,7 +126,8 @@ def _pairs2connection(*pairs, **connection):
 def _db(connection):
     db = connection.pop('db', None)
     if db is None:
-        db = _database(connection.pop('database', None))
+        db = connection.pop('database', None)
+    db = _database(db)
     return db
     
 def get_cstr(*pairs, **connection):
@@ -145,16 +183,6 @@ def get_engine(*pairs, **connection):
     """
     return _get_engine(*pairs, **connection)
     
-_types = {str: String, 'str' : String, 
-          int : Integer, 'int' : Integer,
-          float: Float, 
-          'dec' : sa.DECIMAL(0), 'dec1' : sa.DECIMAL(1), 'dec2' : sa.DECIMAL(2), 'dec3' : sa.DECIMAL(3), 'dec4' : sa.DECIMAL(4), 'dec5' : sa.DECIMAL(5), 
-          datetime.date: DATE, 'date' : DATE,
-          datetime.datetime : DATETIME, 'datetime' : DATETIME,
-          datetime.time: TIME, 'time' : TIME,
-          bin : sa.VARBINARY}
-
-_orders = {1 : asc, True: asc, 'asc': asc, asc : asc, -1: desc, False: desc, 'desc': desc, desc: desc}
 
 @cache
 def _get_table(table_name, schema, db, server):
@@ -314,7 +342,7 @@ def sql_table(table, db = None, non_null = None, nullable = None, _id = None, sc
         meta.create_all(e)
     res = sql_cursor(table = tbl, schema = schema, db = db, server = server, engine = e, 
                      spec = None, selection = None, reader = reader, writer = writer, 
-                     pk = pk, doc = doc)
+                     pk = list(pk) if isinstance(pk, dict) else pk, doc = doc)
     return res
 
 class sql_cursor(object):
@@ -599,6 +627,7 @@ class sql_cursor(object):
         return [c.name for c in self.table.columns if c.server_default is not None]
     
     def _and(self, doc, keys):
+        keys = self._col(keys)
         if len(keys) == 1:
             key = keys[0]
             return self.table.c[key] == doc[key]
@@ -646,6 +675,7 @@ class sql_cursor(object):
         >>> assert t._c(expression) == sa.and_(t.c.a == 1, t.c.b == 2)
         """
         if isinstance(expression, dict):
+            expression = self._col(expression)
             t = self.table.c    
             return sa.and_(*[sa.or_(*[t[k] == i for i in v]) if isinstance(v, list) else t[k] == self._c(v) for k,v in expression.items()]) 
         elif isinstance(expression, (list, tuple)):
@@ -667,25 +697,25 @@ class sql_cursor(object):
         >>> assert t > dict(weight = 30, height = 20) == sa.and_(t.c.weight >= 30, t.c.height > 20)
         """
         if not isinstance(kwargs, dict) and len(self._ids) == 1:
-            kwargs = {self._ids : kwargs}
+            kwargs = self._col({self._ids : kwargs})
         c = self.table.c
         return self.inc(sa.and_(*[c[key] >= value for key, value in kwargs.items()]))
 
     def __gt__(self, kwargs):
         if not isinstance(kwargs, dict) and len(self._ids) == 1:
-            kwargs = {self._ids : kwargs}
+            kwargs = self._col({self._ids : kwargs})
         c = self.table.c
         return self.inc(sa.and_(*[c[key] > value for key, value in kwargs.items()]))
 
     def __le__(self, kwargs):
         if not isinstance(kwargs, dict) and len(self._ids) == 1:
-            kwargs = {self._ids : kwargs}
+            kwargs = self._col({self._ids : kwargs})
         c = self.table.c
         return self.inc(sa.and_(*[c[key] <= value for key, value in kwargs.items()]))
 
     def __lt__(self, kwargs):
         if not isinstance(kwargs, dict) and len(self._ids) == 1:
-            kwargs = {self._ids : kwargs}
+            kwargs = self._col({self._ids : kwargs})
         c = self.table.c
         return self.inc(sa.and_(*[c[key] < value for key, value in kwargs.items()]))
             
@@ -760,7 +790,7 @@ class sql_cursor(object):
             raise ValueError('cannot subtract these columns while the selection is non empty, use self.select() to reset selection')
     
     def drop(self):
-        logger.info('dropping table: %s.%s.%s'%(self.db, 
+        logger.info('dropping table: %s.%s.%s'%(_database(self.db), 
                                                   self.schema, 
                                                   self.table.name))
 
@@ -817,6 +847,19 @@ class sql_cursor(object):
     
     count = __len__
 
+
+    def _col(self, column):
+        columns = self.columns
+        cols = dict(zip(lower(columns), columns))
+        if is_str(column):
+            return cols[column.lower()]
+        elif isinstance(column, (list, tuple)):
+            return [cols[c.lower()] for c in column]
+        elif isinstance(column, dict):
+            return type(column)({cols[k.lower()] : v for k, v in column.items()})
+        else:
+            raise ValueError('column %s cannot be converted to table columns'%column)
+
     @property    
     def columns(self):
         return ulist([col.name for col in self.table.columns])
@@ -825,7 +868,7 @@ class sql_cursor(object):
         if len(value) == 0:
             return self
         res = self.copy()
-        res.selection = as_list(value)
+        res.selection = self._col(as_list(value))
         return res
     
     def _enrich(self, doc, columns = None):
@@ -917,7 +960,9 @@ class sql_cursor(object):
             return edoc
         else:
             edoc = {key: value for key, value in edoc.items() if key in columns}
-            edoc[self.doc] = doc
+            pk = self._pk + [self.doc]
+            drop = [key for key in edoc if key not in pk]
+            edoc[self.doc] = type(doc)({k : v for k, v in doc.items() if k not in drop})
             return edoc
 
     def _writer(self, writer = None, doc = None, kwargs = None):
@@ -999,15 +1044,67 @@ class sql_cursor(object):
         if (is_int(start) and start < 0) or (is_int(stop) and stop < 0):
             n = len(self)
             start = n + start if is_int(start) and start < 0 else start                            
-            stop = n + stop if is_int(stop) and start < 0 else stop
+            stop = n + stop if is_int(stop) and stop < 0 else stop
         if start and self.order is not None:
             statement = statement.offset(start)
             stop = stop if stop is None else stop - start
         if stop is not None:
             statement = statement.limit(1+stop)
         res = list(self.engine.connect().execute(statement))
-        rows = res[slice(start, stop, step)]
-        return rows
+        if start is not None or stop is not None or step is not None:
+            return res[slice(start, stop, step)]
+        else:
+            return res
+    
+    def df(self, decimal2float = True):
+        """
+        This is a more optimized, faster version for reading the table. 
+        It retuns the data as a pd.DataFrame,
+        In addition, it converts NUMERIC type (which is cast to decimal.Decimal) into a float
+
+        
+        Parameters
+        ----------
+        decimal2float: bool
+            converts sql.types.NUMERIC columns into float
+
+        Returns
+        -------
+        pd.DataFrame
+            The data, optimized as a dataframe.
+
+        
+        Example: speed comparison
+        --------
+
+        >>> from pyg import * 
+        >>> t = sql_table(db = 'db', table = 'tbl', nullable = dict(a = int, b = float, c = str))
+        >>> rs =dictable(a = range(100)) * dict(b = list(np.arange(0,100,1.0))) * dict(c = list('abcdefghij'))
+        >>> t = t.insert_many(rs)      
+        
+        ## reading the data as t[::] is slow...
+        >>> x = timer(lambda : t[::])()
+        2022-08-21 18:03:52,594 - pyg - INFO - TIMER: took 0:00:13.139333 sec
+ 
+        ## reading the data as t.df() is much faster...
+        >>> y = timer(lambda : t.df())()
+        2022-08-21 18:04:24,809 - pyg - INFO - TIMER: took 0:00:00.456988 sec
+
+
+        """
+        statement = self.statement()
+        res = self.engine.connect().execute(statement)
+        if _pd_is_old:
+            res = pd.DataFrame(list(res), columns = as_list(self.selection) or self.columns)
+        else:
+            res = pd.DataFrame(res)
+        if decimal2float:
+            for col in res.columns:
+                t = self.table.columns[col].type
+                if isinstance(t, NUMERIC) and not isinstance(t, (FLOAT, INT)):
+                    res[col] = res[col].astype(float)
+        return res
+            
     
     def _rows_to_docs(self, rows, reader = None, load = True):
         """
@@ -1287,6 +1384,7 @@ class sql_cursor(object):
         """
         if len(keys) == 0 and self.selection is not None:
             keys = as_list(self.selection)
+        keys = self._col(keys)
         session = Session(self.engine)
         cols = [self.table.columns[k] for k in keys]
         query = session.query(*cols)
@@ -1305,7 +1403,8 @@ class sql_cursor(object):
         for k,v in params.items():
             text = text.replace(':'+k, '"%s"'%v if is_str(v) else dt2str(v) if is_date(v) else str(v))
             
-        res = 'sql_cursor: %(db)s.%(table)s%(pk)s %(doc)s %(w)s\n%(statement)s\n%(n)i records'%dict(db = self.db, table = prefix, doc = 'DOCSTORE[%s]'%self.doc if self.doc else '', 
+        res = 'sql_cursor: %(db)s.%(table)s%(pk)s %(doc)s %(w)s\n%(statement)s\n%(n)i records'%dict(db = _database(self.db), table = prefix, 
+                                                                                              doc = 'DOCSTORE[%s]'%self.doc if self.doc else '', 
                                                                                               w = '\nwriter: %s\n'%self.writer if is_str(self.writer) else '',
                                                                                               pk = self._pk if self._pk else '', 
                                                                                               n = len(self), 
@@ -1345,6 +1444,6 @@ class sql_cursor(object):
             A unique combination of the server address, db name and table name, identifying the table uniquely. This allows us to create an in-memory representation of the data in pyg-cell
 
         """
-        return ('server', self.server or get_server()), ('db', self.db), ('schema', self.schema), ('table', self.table.name)
+        return ('server', self.server), ('db', self.db), ('schema', self.schema), ('table', self.table.name)
 
 
